/* 
 * kpci_pio96.c -- Linux device driver for
 *     Keithley KPCI-PIO96 digital I/O card.
 * Will work on x86 platform only.
 * Author: Igor Volobouev, igv@lbl.gov  06/03/2002
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <asm/system.h>

#if LINUX_VERSION_CODE >= 0x020100
#if LINUX_VERSION_CODE < 0x020200
#error Linux kernels 2.1.x are not supported. Please switch to a stable kernel.
#endif
#endif

#if LINUX_VERSION_CODE >= 0x020300
#if LINUX_VERSION_CODE < 0x020400
#error Linux kernels 2.3.x are not supported. Please switch to a stable kernel.
#endif
#endif

#if LINUX_VERSION_CODE < 0x020200
#include <linux/bios32.h>
#endif

#ifndef PCI_VENDOR_ID_KEITHLEY
#define PCI_VENDOR_ID_KEITHLEY 0x11f3
#endif

#ifndef PCI_DEVICE_ID_KPCI_PIO96
#define PCI_DEVICE_ID_KPCI_PIO96 0x12
#endif

#ifndef KPCI_PIO96_MAJOR_DEV
#define KPCI_PIO96_MAJOR_DEV 26    /* Major device number */
#endif

#define KPCI_MAX_DEVICES 0x100

/* Comment out the next #define if you want to access
   the card from several processes simultaneously */
#define ENABLE_BUSY_CONTROL

static const char insmod_info[] = "KPCI-PIO96 driver v0.5, igv@lbl.gov";
static int major_number = KPCI_PIO96_MAJOR_DEV;

typedef struct {
    unsigned int phys_addr;
    int open;
} Kpci_device;

static Kpci_device *kpci_devices = NULL;
static int n_devices = 0;

static void kpci_vma_open(struct vm_area_struct *area)
{
    MOD_INC_USE_COUNT;
}

static void kpci_vma_close(struct vm_area_struct *area)
{
    MOD_DEC_USE_COUNT;
}

#if LINUX_VERSION_CODE >= 0x020400
static struct page *
#else
static unsigned long 
#endif
kpci_vma_nopage(struct vm_area_struct *area,
		unsigned long address,
		int write_access)
{
    /* This function may be called by "mremap" system call.
     * We aren't really going to support remapping. Instead,
     * make sure that the current process receives SIGBUS.
     */
    return 0;
}

static struct vm_operations_struct kpci_vm_ops = {
    open:    kpci_vma_open,
    close:   kpci_vma_close,
    nopage:  kpci_vma_nopage,
};

static int kpci_open(struct inode *inode, struct file *fileptr)
{
    int minor;
#ifdef ENABLE_BUSY_CONTROL
    unsigned long flags;
#endif

#ifdef DEBUG
    printk(KERN_INFO "kpci_pio96: open\n");
#endif
    minor = MINOR(inode->i_rdev);
    if (minor >= n_devices)
	return -ENODEV;
    
#ifdef ENABLE_BUSY_CONTROL
    save_flags(flags);
    cli();
    if (kpci_devices[minor].open != 0) {
	restore_flags(flags);
#ifdef DEBUG
	printk(KERN_INFO "kpci_pio96: device busy\n");
#endif
	return -EBUSY;
    } else {
	kpci_devices[minor].open = 1;
	restore_flags(flags);
    }
#else
    kpci_devices[minor].open++;
#endif
    
    MOD_INC_USE_COUNT;
    return 0;
}

#if LINUX_VERSION_CODE >= 0x020200
static int kpci_close(struct inode *inode, struct file *fileptr)
#else
static void kpci_close(struct inode *inode, struct file *fileptr)
#endif
{
#ifdef DEBUG
    printk(KERN_INFO "kpci_pio96: close\n");
#endif
    kpci_devices[MINOR(inode->i_rdev)].open--;
    MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE >= 0x020200
    return 0;
#endif
}

#if LINUX_VERSION_CODE < 0x020200
static int kpci_mmap(struct inode *inode, struct file *fileptr,
		     struct vm_area_struct *vma)
#else
static int kpci_mmap(struct file *fileptr, struct vm_area_struct *vma)
#endif
{
    unsigned long physical_addr, size;

#if LINUX_VERSION_CODE >= 0x020200
    struct inode *inode = fileptr->f_dentry->d_inode;
#endif
	
    if (vma->vm_ops) {
	printk ("kpci_pio96: kpci_mmap: vm_ops alredy initialized\n");
	return -EINVAL;
    }
#if LINUX_VERSION_CODE >= 0x020400
    if (vma->vm_pgoff) {
#else
    if (vma->vm_offset) {
#endif
	printk ("kpci_pio96: kpci_mmap: offset must be 0\n");
	return -EINVAL;
    }
    size = vma->vm_end - vma->vm_start;
    if (size != PAGE_SIZE) {
	printk ("kpci_pio96: kpci_mmap: invalid vma size\n");
	return -EINVAL;
    }
    
    physical_addr = (unsigned long)kpci_devices[MINOR(inode->i_rdev)].phys_addr;

#ifdef DEBUG
    printk(KERN_INFO "kpci_pio96: vma->vm_start is 0x%08lx\n",
	   vma->vm_start);
    printk(KERN_INFO "kpci_pio96: physical address is 0x%08lx\n", 
	   physical_addr);
    printk(KERN_INFO "kpci_pio96: size is 0x%08lx\n", size);
#endif
    
    if (remap_page_range(vma->vm_start, physical_addr,
			 size, vma->vm_page_prot))
	return -EAGAIN;
    
    vma->vm_ops = &kpci_vm_ops;
#if LINUX_VERSION_CODE >= 0x020200
    vma->vm_file = fileptr;
#else
    vma->vm_inode = inode;
#endif
#if LINUX_VERSION_CODE >= 0x020400
    atomic_inc(&inode->i_count);
#else
    inode->i_count++;
#endif

    MOD_INC_USE_COUNT;
    return 0;
}

static struct file_operations kpci_pio96_fops = {
    mmap:     kpci_mmap,
    open:     kpci_open,
    release:  kpci_close,
};

int init_module(void) 
{
#ifdef CONFIG_PCI
    int i, status;

#if LINUX_VERSION_CODE < 0x020200
    unsigned char bus[KPCI_MAX_DEVICES], function[KPCI_MAX_DEVICES];
#else
    struct pci_dev *dev = NULL;
    struct pci_dev *devarray[KPCI_MAX_DEVICES];
#endif

    /* Check for PCI support in the computer hardware */
#if LINUX_VERSION_CODE < 0x020200
    if (!pcibios_present())
#else
    if (!pci_present())
#endif
	return -ENODEV;
    
    /* Find out how many cards we have in the system */
#if LINUX_VERSION_CODE < 0x020200
    for (n_devices = 0; n_devices < KPCI_MAX_DEVICES; n_devices++) {
	status = pcibios_find_device(PCI_VENDOR_ID_KEITHLEY, PCI_DEVICE_ID_KPCI_PIO96,
				     n_devices, bus+n_devices, function+n_devices);
	if (status != PCIBIOS_SUCCESSFUL)
	    break;
    }

    /* When all cards have been found, "pcibios_find_device" should
       return PCIBIOS_DEVICE_NOT_FOUND. */
    if (status != PCIBIOS_DEVICE_NOT_FOUND) {
	printk(KERN_WARNING "kpci_pio96: pci error: %s\n", 
	       pcibios_strerror(status));
	return -EIO;
    }
#else
    n_devices = 0;
    while (n_devices < KPCI_MAX_DEVICES && 
	   (dev = pci_find_device(PCI_VENDOR_ID_KEITHLEY, PCI_DEVICE_ID_KPCI_PIO96, dev)))
	devarray[n_devices++] = dev;
#endif

    if (n_devices == 0) {
	printk(KERN_WARNING "kpci_pio96: I/O card not found\n");
	return -ENODEV;
    }

    if (n_devices == KPCI_MAX_DEVICES) {
	printk(KERN_WARNING "kpci_pio96: reached max. number of cards ( %d )\n",
	       KPCI_MAX_DEVICES);
    }

#ifdef DEBUG
    printk(KERN_INFO "kpci_pio96: found %d card(s)\n", n_devices);
#endif

    kpci_devices = (Kpci_device *)kmalloc(n_devices*sizeof(Kpci_device), GFP_KERNEL);
    if (kpci_devices == NULL)
	return -ENOMEM;
    else
	memset(kpci_devices, 0, n_devices*sizeof(Kpci_device));

    /* Now, look up the physical addresses */
    for (i = 0; i < n_devices; i++) {
#if LINUX_VERSION_CODE < 0x020200
	status = pcibios_read_config_dword(bus[i], function[i],
					   PCI_BASE_ADDRESS_1,
					   &kpci_devices[i].phys_addr);
	if (status) {
	    printk(KERN_WARNING "kpci_pio96: pci error: %s\n", 
		   pcibios_strerror(status));
	    kfree(kpci_devices);
	    kpci_devices = NULL;
	    return -EIO;	    
	}
#else
	status = pcibios_read_config_dword(devarray[i]->bus->number,
					   devarray[i]->devfn, 
					   PCI_BASE_ADDRESS_1, 
					   &kpci_devices[i].phys_addr);
	if (status) {
	    printk(KERN_WARNING "kpci_pio96: pci error %d\n", status);
	    /* Where the hell did the "pcibios_strerror" disappear? */
	    kfree(kpci_devices);
	    kpci_devices = NULL;
	    return -EIO;	    
	}
#endif
    }

    status = register_chrdev(major_number, "kpci_pio96", &kpci_pio96_fops);
    if (status) {
	printk(KERN_WARNING "kpci_pio96: failed to get major %d\n", major_number);
	kfree(kpci_devices);
	kpci_devices = NULL;
	return status;
    } else {
	printk(KERN_INFO "%s\n", insmod_info);
    }

    return 0;

#else
    /* There is no support for PCI BIOS calls in the kernel */
    return -ENODEV;
#endif /* CONFIG_PCI*/
}

void cleanup_module(void) 
{
#ifdef DEBUG
    printk(KERN_INFO "kpci_pio96: unloading\n");
#endif
    if (unregister_chrdev(major_number, "kpci_pio96") != 0) {
	printk(KERN_WARNING "kpci_pio96: cleanup_module failed\n");
    }
    if (kpci_devices) {
	kfree(kpci_devices);
	kpci_devices = NULL;
    }
}
