#ifndef KPCI_PIO96_H
#define KPCI_PIO96_H

/* Number of port groups on the KPCI-PIO96 card */
#define KPCI_PORTGROUPS 4

/* The following structure describes KPCI-PIO96 ports and
 * control registers:
 */
typedef struct {
    unsigned int port[KPCI_PORTGROUPS];
    unsigned int ctrl[KPCI_PORTGROUPS];
} Kpci_registers;

/* Masks for the control register bits. The ports work for
 * input when their direction bits are set to 0 and for
 * output when these bits are set to 1
 */
#define KPCI_PORTC_HI_OUT 0x8
#define KPCI_PORTC_LO_OUT 0x4
#define KPCI_PORTB_OUT 0x2
#define KPCI_PORTA_OUT 0x1

/* Port to the I/O connector mapping:
 *
 *  KPCI-PIO96        KPCI-PIO96
 *  I/O connector     port/bit  
 *  pin number    
 *      1               A1/7
 *      3               A1/6
 *      5               A1/5
 *      7               A1/4
 *      9               A1/3
 *      11              A1/2
 *      13              A1/1
 *      15              A1/0
 *      17              C1/7
 *      19              C1/6
 *      21              C1/5
 *      23              C1/4
 *      25              C1/3
 *      27              C1/2
 *      29              C1/1
 *      31              C1/0
 *      33              B1/7
 *      35              B1/6
 *      37              B1/5
 *      39              B1/4
 *      41              B1/3
 *      43              B1/2
 *      45              B1/1
 *      47              B1/0
 *
 * All even-numbered pins are grounded. Pin number 49 is routed
 * to the +5V line of the PCI bus.
 *
 * 32-bit port access works as follows:
 *   bits 0:7    -   PA_0:PA_7
 *   bits 8:15   -   PB_0:PB_7
 *   bits 16:19  -   PC_0:PC_3
 *   bits 20:27  -   undefined
 *   bits 28:31  -   PC_4:PC_7
 *
 * The following macros are provided in order to deal with
 * the gap at bits 20:27.
 */

#define kpci_toport32(x) (((x) & 0xfffff) | \
			  (((x) & 0xf00000) << 8))

#define kpci_fromport32(tmp) (((tmp) & 0xfffff) | \
			      (((unsigned)(tmp) >> 8) & 0xf00000))

#define kpci_toport(a, b, c) (((a) & 0xff) | \
			      (((b) & 0xff) << 8) | \
			      (((c) & 0xf) << 16) | \
			      (((c) & 0xf0) << 24))

#define kpci_fromport(tmp, a, b, c) do {\
    a = (tmp) & 0xff; \
    b = ((tmp) & 0xff00) >> 8; \
    c = ((tmp) & 0xf0000) >> 16 | \
        (unsigned)((tmp) & 0xf0000000) >> 24;} while(0)

/* Macros "kpci_fromport" and "kpci_fromport32" should not be
 * used directly with the mapped ports because the PCI bus
 * access is very slow compared to the memory access. Instead,
 * these macros should act upon some temporary variable written
 * with the port value.
 */

#endif /* not KPCI_PIO96_H */
